from django.db import models

from sitecats.models import ModelWithCategory


class Comment(ModelWithCategory):

    title = models.CharField('title', max_length=255)


class Article(ModelWithCategory):

    title = models.CharField('title', max_length=255)

    def get_category_absolute_url(self, category):
        return '%s/%s' % (category.id, self.title)



from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class SitecatsConfig(AppConfig):
    """Sitecats configuration."""

    name = 'sitecats'
    verbose_name = _('Site Categories')

    _cat_cache = None

    def get_categories_cache(self):
        return self._cat_cache

    def ready(self):
        """Instantiate global cache object when ready.

        :return:
        """
        from sitecats.utils import Cache
        self._cat_cache = Cache()



from inspect import getargspec
from collections import namedtuple, OrderedDict, Callable

from django.utils.translation import ugettext_lazy as _, ungettext_lazy
from django.contrib.contenttypes.models import ContentType
from django.contrib import messages

from .settings import UNRESOLVED_URL_MARKER
from .utils import get_category_model, get_tie_model, get_cache
from .exceptions import SitecatsConfigurationError, SitecatsSecurityException, SitecatsNewCategoryException, \
    SitecatsValidationError


def get_category_aliases_under(parent_alias=None):
    """Returns a list of category aliases under the given parent.

    Could be useful to pass to `ModelWithCategory.enable_category_lists_editor`
    in `additional_parents_aliases` parameter.

    :param str|None parent_alias: Parent alias or None to categories under root
    :rtype: list
    :return: a list of category aliases
    """
    return [ch.alias for ch in get_cache().get_children_for(parent_alias, only_with_aliases=True)]


def get_category_lists(init_kwargs=None, additional_parents_aliases=None, obj=None):
    """Returns a list of CategoryList objects, optionally associated with
    a given model instance.

    :param dict|None init_kwargs:
    :param list|None additional_parents_aliases:
    :param Model|None obj: Model instance to get categories for
    :rtype: list
    :return:
    """
    init_kwargs = init_kwargs or {}
    additional_parents_aliases = additional_parents_aliases or []

    parent_aliases = additional_parents_aliases
    if obj is not None:
        ctype = ContentType.objects.get_for_model(obj)
        cat_ids = [
            item[0] for item in
            get_tie_model().objects.filter(content_type=ctype, object_id=obj.id).values_list('category_id').all()
        ]
        parent_aliases = list(get_cache().get_parents_for(cat_ids).union(additional_parents_aliases))
    lists = []

    aliases = get_cache().sort_aliases(parent_aliases)
    categories_cache = get_cache().get_categories(aliases, obj)

    for parent_alias in aliases:
        catlist = CategoryList(parent_alias, **init_kwargs)  # TODO Burned in class name. Make more customizable.
        if obj is not None:
            catlist.set_obj(obj)
        # Optimization. To get DB hits down.
        cache = []
        try:
            cache = categories_cache[parent_alias]
        except KeyError:
            pass
        catlist.set_get_categories_cache(cache)
        lists.append(catlist)

    return lists


class CategoryList(object):
    """Represents a set on categories under a parent category on page."""
    _cache_category = None
    _cache_get_categories = None

    #TODO custom template

    def __init__(self, alias=None, show_title=False, show_links=True, cat_html_class=''):
        """
        :param str alias: Alias of a category to construct a list from (list will include subcategories)

        :param bool show_title: Flag to render parent category title

        :param bool|callable show_links: Boolean flag to render links for category pages,
            or a callable which accepts Category instance and returns an URL for it.
            If boolean and True links will be set to UNRESOLVED_URL_MARKER (useful
            for client-side links generation based on data-* attrs of HTML elements).

        :param str cat_html_class: HTML classes to be added to categories

        :return:
        """
        self.alias = alias
        self.show_title = show_title
        self._url_resolver = None
        if isinstance(show_links, Callable):
            self._url_resolver = show_links
            show_links = True
        self.show_links = show_links
        self.cat_html_class = cat_html_class
        self.obj = None
        self.editor = None

    def __str__(self):
        """Returns alias.

        :rtype: str
        :return: alias
        """
        s = self.alias
        if s is None:
            s = ''
        return s

    def set_get_categories_cache(self, val):
        """Sets prefetched data to be returned by `get_categories()` later on.

        :param list val:
        :return:
        """
        self._cache_get_categories = val

    def get_category_url(self, category):
        """Returns URL for a given Category object from this list.

         First tries to get it with a callable passed as `show_links` init param of this list.
         Secondly tries to get it with `get_category_absolute_url` method of an object associated with this list.

        :param Category category:
        :rtype: str
        :return: URL
        """
        if self._url_resolver is not None:
            return self._url_resolver(category)
        if self.obj and hasattr(self.obj, 'get_category_absolute_url'):
            return self.obj.get_category_absolute_url(category)
        return UNRESOLVED_URL_MARKER

    def set_obj(self, obj):
        """Sets a target object for categories to be filtered upon.

        `ModelWithCategory` heir is expected.

        If not set CategoryList will render actual categories.
        If set CategoryList will render just object-to-categories ties.

        :param obj: `ModelWithCategory` heir
        """
        self.obj = obj

    def enable_editor(self, allow_add=True, allow_remove=False, allow_new=False, min_num=None, max_num=None,
                      render_button=True, category_separator=None, show_category_choices=True):
        """Enables editor controls for this category list.

        :param bool allow_add: Flag to allow adding object-to-categories ties
        :param bool allow_remove: Flag to allow remove of object-to-categories ties or categories themselves
        :param bool allow_new: Flag to allow new categories creation
        :param None|int min_num: Child items minimum for this list
            (object-to-categories ties or categories themselves)
        :param None|int max_num: Child items maximum for this list
            (object-to-categories ties or categories themselves)
        :param bool render_button: Flag to allow buttons rendering for forms of this list
        :param str|None category_separator: String to consider it a category separator.
        :param bool show_category_choices: Flag to render a choice list of available subcategories
            for each CategoryList
        """
        # DRY: translate method args into namedtuple args.
        args, n, n, n = getargspec(self.enable_editor)
        locals_ = locals()
        self.editor = namedtuple('CategoryEditor', args)(**{arg: locals_[arg] for arg in args})

    def get_category_model(self):
        """Returns category model for this list (parent category for categories in the list) or None.

        :return: CategoryModel|None
        """
        if self._cache_category is None:
            self._cache_category = get_cache().get_category_by_alias(self.alias)
        return self._cache_category

    def get_category_attr(self, name, default=False):
        """Returns a custom attribute of a category model for this list.

        :param str name: Attribute name
        :param object default: Default value if attribute is not found
        :return: attribute value
        """
        category = self.get_category_model()
        return getattr(category, name, default)

    def get_id(self):
        """Returns ID attribute of a category of this list.

        :rtype: int|None
        :return: id
        """
        return self.get_category_attr('id', None)

    def get_title(self):
        """Returns `title` attribute of a category of this list.

        :rtype: str
        :return: title
        """
        return self.get_category_attr('title', _('Categories'))

    def get_note(self):
        """Returns `note` attribute of a category of this list.

        :rtype: str
        :return: note
        """
        return self.get_category_attr('note', '')

    def get_categories(self, tied_only=None):
        """Returns a list of actual subcategories.

        :param bool tied_only: Flag to get only categories with ties. Ties stats are stored in `ties_num` attrs.
        :rtype: list
        :return: a list of actual subcategories
        """
        if self._cache_get_categories is not None:
            return self._cache_get_categories
        if tied_only is None:
            tied_only = self.obj is not None
        return get_cache().get_categories(self.alias, self.obj, tied_only=tied_only)

    def get_choices(self):
        """Returns available subcategories choices list.

        :rtype: list
        :return: list of Category objects
        """
        return get_cache().get_children_for(self.alias)


class CategoryRequestHandler(object):
    """This one can handle requests issued by CategoryList editors. Can be used in views."""

    list_cls = CategoryList  # For customization purposes.
    KNOWN_ACTIONS = ('add', 'remove')

    def __init__(self, request, obj=None, error_messages_extra_tags=None):
        """
        :param Request request: Django request object
        :param Model obj: `ModelWithCategory` heir to bind CategoryList objects upon.
        """
        self._request = request
        self._lists = OrderedDict()
        self._obj = obj
        self.error_messages_extra_tags = error_messages_extra_tags or ''

    def register_lists(self, category_lists, lists_init_kwargs=None, editor_init_kwargs=None):
        """Registers CategoryList objects to handle their requests.

        :param list category_lists: CategoryList objects
        :param dict lists_init_kwargs: Attributes to apply to each of CategoryList objects
        """
        lists_init_kwargs = lists_init_kwargs or {}
        editor_init_kwargs = editor_init_kwargs or {}

        for lst in category_lists:
            if isinstance(lst, str):  # Spawn CategoryList object from base category alias.
                lst = self.list_cls(lst, **lists_init_kwargs)
            elif not isinstance(lst, CategoryList):
                raise SitecatsConfigurationError(
                    '`CategoryRequestHandler.register_lists()` accepts only '
                    '`CategoryList` objects or category aliases.'
                )

            if self._obj:
                lst.set_obj(self._obj)

            for name, val in lists_init_kwargs.items():  # Setting CategoryList attributes from kwargs.
                setattr(lst, name, val)

            lst.enable_editor(**editor_init_kwargs)

            self._lists[lst.get_id()] = lst

    @classmethod
    def action_remove(cls, request, category_list):
        """Handles `remove` action from CategoryList editor.

        Removes an actual category if a target object is not set for the list.
        Removes a tie-to-category object if a target object is set for the list.

        :param Request request: Django request object
        :param CategoryList category_list: CategoryList object to operate upon.
        :return: True on success otherwise and exception from SitecatsException family is raised.
        """
        if not category_list.editor.allow_remove:
            raise SitecatsSecurityException(
                '`action_remove()` is not supported by parent `%s`category.' % category_list.alias)

        category_id = int(request.POST.get('category_id', 0))
        if not category_id:
            raise SitecatsSecurityException(
                'Unsupported `category_id` value - `%s` - is passed to `action_remove()`.' % category_id)

        category = get_cache().get_category_by_id(category_id)
        if not category:
            raise SitecatsSecurityException('Unable to get `%s` category in `action_remove()`.' % category_id)

        cat_ident = category.alias or category.id

        if category.is_locked:
            raise SitecatsSecurityException('`action_remove()` is not supported by `%s` category.' % cat_ident)

        if category.parent_id != category_list.get_id():
            raise SitecatsSecurityException(
                '`action_remove()` is unable to remove `%s`: '
                'not a child of parent `%s` category.' % (cat_ident, category_list.alias)
            )

        min_num = category_list.editor.min_num

        def check_min_num(num):
            if min_num is not None and num-1 < min_num:
                subcats_str = ungettext_lazy('subcategory', 'subcategories', min_num)
                error_msg = _(
                    'Unable to remove "%(target_category)s" category from "%(parent_category)s": '
                    'parent category requires at least %(num)s %(subcats_str)s.'
                ) % {
                    'target_category': category.title,
                    'parent_category': category_list.get_title(),
                    'num': min_num,
                    'subcats_str': subcats_str
                }
                raise SitecatsValidationError(error_msg)

        child_ids = get_cache().get_child_ids(category_list.alias)
        check_min_num(len(child_ids))
        if category_list.obj is None:  # Remove category itself and children.
            category.delete()
        else:  # Remove just a category-to-object tie.
            # TODO filter user/status
            check_min_num(category_list.obj.get_ties_for_categories_qs(child_ids).count())
            category_list.obj.remove_from_category(category)

        return True

    @classmethod
    def action_add(cls, request, category_list):
        """Handles `add` action from CategoryList editor.

        Adds an actual category if a target object is not set for the list.
        Adds a tie-to-category object if a target object is set for the list.

        :param Request request: Django request object
        :param CategoryList category_list: CategoryList object to operate upon.
        :return: CategoryModel object on success otherwise and exception from SitecatsException family is raised.
        """
        if not category_list.editor.allow_add:
            raise SitecatsSecurityException('`action_add()` is not supported by `%s` category.' % category_list.alias)

        titles = request.POST.get('category_title', '').strip()
        if not titles:
            raise SitecatsSecurityException(
                'Unsupported `category_title` value - `%s` - is passed to `action_add()`.' % titles)

        if category_list.editor.category_separator is None:
            titles = [titles]
        else:
            titles = [
                title.strip() for title in titles.split(category_list.editor.category_separator) if title.strip()
            ]

        def check_max_num(num, max_num, category_title):
            if max_num is not None and num+1 > max_num:
                subcats_str = ungettext_lazy('subcategory', 'subcategories', max_num)
                error_msg = _(
                    'Unable to add "%(target_category)s" category into "%(parent_category)s": '
                    'parent category can have at most %(num)s %(subcats_str)s.'
                ) % {
                    'target_category': category_title,
                    'parent_category': category_list.get_title(),
                    'num': max_num,
                    'subcats_str': subcats_str
                }
                raise SitecatsValidationError(error_msg)

        target_category = None
        for category_title in titles:
            exists = get_cache().find_category(category_list.alias, category_title)

            if exists and category_list.obj is None:  # Already exists.
                return exists

            if not exists and not category_list.editor.allow_new:
                error_msg = _(
                    'Unable to create a new "%(new_category)s" category inside of "%(parent_category)s": '
                    'parent category does not support this action.'
                ) % {
                    'new_category': category_title,
                    'parent_category': category_list.get_title()
                }
                raise SitecatsNewCategoryException(error_msg)

            max_num = category_list.editor.max_num
            child_ids = get_cache().get_child_ids(category_list.alias)
            if not exists:  # Add new category.
                if category_list.obj is None:
                    check_max_num(len(child_ids), max_num, category_title)
                # TODO status
                target_category = get_category_model().add(
                    category_title, request.user, parent=category_list.get_category_model()
                )
            else:
                target_category = exists  # Use existing one for a tie.

            if category_list.obj is not None:
                # TODO status
                check_max_num(category_list.obj.get_ties_for_categories_qs(child_ids).count(), max_num, category_title)
                category_list.obj.add_to_category(target_category, request.user)

        return target_category

    def listen(self):
        """Instructs handler to listen to Django request and handle
        CategoryList editor requests (if any).

        :return: None on success otherwise and exception from SitecatsException family is raised.
        """
        requested_action = self._request.POST.get('category_action', False)

        if not requested_action:
            return None  # No action supplied. Pass.

        if requested_action not in self.KNOWN_ACTIONS:
            raise SitecatsSecurityException('Unknown `category_action` - `%s` - requested.')

        category_base_id = self._request.POST.get('category_base_id', False)
        if category_base_id == 'None':
            category_base_id = None
        else:
            category_base_id = int(category_base_id)
        if category_base_id not in self._lists.keys():
            raise SitecatsSecurityException('Unknown `category_base_id` - `%s` - requested.')

        category_list = self._lists[category_base_id]
        if category_list.editor is None:
            raise SitecatsSecurityException('Editor is disabled for `%s` category.' % category_list.alias)

        action_method = getattr(self, 'action_%s' % requested_action)

        try:
            return action_method(self._request, category_list)
        except SitecatsNewCategoryException as e:
            messages.error(self._request, e, extra_tags=self.error_messages_extra_tags, fail_silently=True)
            return None
        except SitecatsValidationError as e:
            messages.error(self._request, e.messages[0], extra_tags=self.error_messages_extra_tags, fail_silently=True)
            return None
        finally:
            self._request.POST = {}  # Prevent other forms fail.

    def get_lists(self):
        """Returns a list of previously registered CategoryList objects.

        :rtype: list
        :return: A list of CategoryList objects.
        """
        return list(self._lists.values())



from uuid import uuid4

import pytest
from django.db.utils import IntegrityError
from django.template.base import TemplateSyntaxError
from django.template.context import Context

from sitecats.exceptions import SitecatsLockedCategoryDelete, SitecatsConfigurationError
from sitecats.models import Category, Tie
from sitecats.settings import UNRESOLVED_URL_MARKER
from sitecats.toolbox import CategoryList, get_category_aliases_under, get_tie_model, \
    get_category_model

MODEL_TIE = get_tie_model()
MODEL_CATEGORY = get_category_model()

from sitecats.tests.testapp.models import Comment, Article


@pytest.fixture
def create_comment():
    from sitecats.tests.testapp.models import Comment

    def create_comment_():
        comment = Comment(title='comment%s' % uuid4().hex)
        comment.save()
        return comment

    return create_comment_


@pytest.fixture
def create_article():
    from sitecats.tests.testapp.models import Article

    def create_article_():
        article = Article(title='article%s' % uuid4().hex)
        article.save()
        return article

    return create_article_


@pytest.fixture
def create_category(user_create):

    def create_category_(title=None, alias=None, parent=None, creator=None, **kwargs):
        title = title or 'cat%s' % uuid4().hex

        creator = creator or user_create()

        cat = Category(title=title, alias=alias, parent=parent, creator=creator, **kwargs)
        cat.save()

        return cat

    return create_category_


class TestTemplateTags(object):

    @pytest.fixture
    def setup(self, user, create_article, create_category):
        self.user = user

        self.cat1 = create_category(alias='cat1')
        self.cl_cat1 = CategoryList('cat1')

        self.cat2 = create_category(alias='cat2')
        self.cl_cat2 = CategoryList('cat2', show_links=lambda category: category.id)

        self.cat3 = create_category(alias='cat3')
        self.cat31 = create_category(alias='cat31', parent=self.cat3)
        self.cl_cat3 = CategoryList('cat3')

        self.art1 = create_article()

    def test_sitecats_url(self, setup, template_render_tag, template_context):

        with pytest.raises(TemplateSyntaxError):
            template_render_tag('sitecats', 'sitecats_url')

        # testing UNRESOLVED_URL_MARKER
        context = template_context({'my_category': self.cat1, 'my_list': self.cl_cat1})
        result = template_render_tag('sitecats', 'sitecats_url for my_category using my_list', context)
        assert result == UNRESOLVED_URL_MARKER

        # testing ModelWithCategory.get_category_absolute_url()
        self.cl_cat1.set_obj(self.art1)
        expected_url = '%s/%s' % (self.cat1.id, self.art1.title)
        assert template_render_tag('sitecats', 'sitecats_url for my_category using my_list', context) == expected_url

        template_render_tag('sitecats', 'sitecats_url for my_category using my_list as someurl', context)
        assert context.get('someurl') == expected_url

        # testing CategoryList.show_links
        context = template_context({'my_category': self.cat2, 'my_list': self.cl_cat2})
        result = template_render_tag('sitecats', 'sitecats_url for my_category using my_list', context)
        assert result == str(self.cat2.id)

    def test_sitecats_categories(self, setup, template_render_tag, template_context, settings):

        with pytest.raises(TemplateSyntaxError):
            template_render_tag('sitecats', 'sitecats_categories')

        # testing CategoryList passed into `from` clause
        context = template_context({'my_categories_list': self.cl_cat3})
        result = template_render_tag('sitecats', 'sitecats_categories from my_categories_list', context)
        print(result)
        assert ('data-catid="%s"' % self.cat3.id) in result
        assert ('data-catid="%s"' % self.cat31.id) in result

        # testing list of CategoryList passed into `from` clause
        context = template_context({'my_categories_list': [self.cl_cat3]})
        result = template_render_tag('sitecats', 'sitecats_categories from my_categories_list', context)
        assert ('data-catid="%s"' % self.cat3.id) in result
        assert ('data-catid="%s"' % self.cat31.id) in result

        # testing unknown type passed into `from` clause
        context = template_context({'my_categories_list': object()})
        result = template_render_tag('sitecats', 'sitecats_categories from my_categories_list', context)
        assert result == ''

        settings.DEBUG = True

        with pytest.raises(SitecatsConfigurationError):
            context = template_context({'my_categories_list': object()})
            template_render_tag('sitecats', 'sitecats_categories from my_categories_list', context)


class TestCategoryModel(object):

    def test_alias_unique(self, create_category):
        create_category(alias='doubled')

        with pytest.raises(IntegrityError):
            create_category(None, 'doubled')

    def test_title_unique_in_parent(self, create_category):
        cat1 = create_category('title1')
        create_category('title1', parent=cat1)

        with pytest.raises(IntegrityError):
            create_category('title1', None, cat1)

    def test_delete(self, create_category):
        cat1 = create_category()
        cat1.delete()

        cat2 = create_category(is_locked=True)

        with pytest.raises(SitecatsLockedCategoryDelete):
            cat2.delete()

        cat2.is_locked = False
        cat2.delete()

    def test_add(self, user, create_category):
        cat1 = Category.add('new', user)

        assert cat1.pk is not None
        assert cat1.creator == user

        parent_cat = create_category()
        cat2 = Category.add('newest', user, parent=parent_cat)

        assert cat2.pk is not None
        assert cat2.creator == user
        assert cat2.parent == parent_cat

    def test_sort_order(self, create_category):
        cat1 = create_category()
        assert cat1.sort_order == cat1.pk

        cat2 = create_category(sort_order=128)
        assert cat2.sort_order == 128

    def test_strip_title(self, create_category):
        cat1 = create_category(' my title ')
        assert cat1.title == 'my title'


class TestTie(object):

    def test_get_linked_objects(self, user, create_article, create_comment, create_category):
        cat1 = create_category()
        cat2 = create_category()
        cat3 = create_category()

        article1 = create_article()
        article2 = create_article()
        article3 = create_article()

        article1.add_to_category(cat1, user)
        article2.add_to_category(cat1, user)
        article3.add_to_category(cat1, user)
        article2.add_to_category(cat2, user)
        article3.add_to_category(cat3, user)

        linked = MODEL_TIE.get_linked_objects()
        keys = linked.keys()
        assert Article in keys
        assert Comment not in keys

        assert article1 in linked[Article]
        assert article2 in linked[Article]
        assert article3 in linked[Article]

        comment1 = create_comment()
        comment1.add_to_category(cat1, user)

        linked = MODEL_TIE.get_linked_objects()
        keys = linked.keys()
        assert Article in keys
        assert Comment in keys

        assert comment1 in linked[Comment]

        # testing `id_only`
        linked = MODEL_TIE.get_linked_objects(id_only=True)
        assert article1.id in linked[Article]
        assert article2.id in linked[Article]
        assert article3.id in linked[Article]
        assert comment1.id in linked[Comment]

        # testing `by_category`
        linked = MODEL_TIE.get_linked_objects(by_category=True)
        keys = linked.keys()
        assert cat1 in keys
        assert cat2 in keys
        assert cat3 in keys

        assert len(linked[cat1]) == 2  # Article + Comment
        assert len(linked[cat2]) == 1
        assert len(linked[cat3]) == 1


class TestModelWithCategory(object):

    # TODO set_category_lists_init_kwargs, get_category_lists, enable_category_lists_editor

    def test_add_to_category(self, user, create_article, create_category):
        cat = create_category()

        article = create_article()
        tie = article.add_to_category(cat, user)

        assert isinstance(tie, Tie)
        assert tie.category == cat
        assert tie.creator == user

    def test_remove_from_category(self, user, create_article, create_category):
        cat = create_category()

        article = create_article()
        article.add_to_category(cat, user)

        ties = Article.get_ties_for_categories_qs([cat])
        assert len(ties) == 1

        article.remove_from_category(cat)
        ties = Article.get_ties_for_categories_qs([cat])
        assert len(ties) == 0

    def test_get_ties_for_categories_qs(self, user_create, create_article, create_category):

        user = user_create()
        user2 = user_create()
        cat1 = create_category()
        cat2 = create_category()

        article = create_article()
        tie1 = article.add_to_category(cat1, user)
        tie2 = article.add_to_category(cat2, user2)

        ties = Article.get_ties_for_categories_qs([cat1, cat2])

        assert len(ties) == 2
        assert tie1 in ties
        assert tie2 in ties

        ties = Article.get_ties_for_categories_qs([cat1, cat2], user=user2)

        assert len(ties) == 1
        assert tie1 not in ties
        assert tie2 in ties

    def test_get_from_category_qs(self, user, create_article, create_comment, create_category):
        cat1 = create_category()

        article = create_article()
        comment = create_comment()

        article.add_to_category(cat1, user)
        comment.add_to_category(cat1, user)

        articles_in_cat = Article.get_from_category_qs(cat1)
        assert len(articles_in_cat) == 1
        assert article in articles_in_cat

        comments_in_cat = Comment.get_from_category_qs(cat1)
        assert len(comments_in_cat) == 1
        assert comment in comments_in_cat


class TestToolbox(object):

    def test_get_category_aliases_under(self, create_category):
        cat1 = create_category(alias='cat1',)
        cat2 = create_category(alias='cat2',)
        cat3 = create_category(alias='cat3')

        cat11 = create_category(alias='cat11', parent=cat1)
        cat111 = create_category(parent=cat1)

        root = get_category_aliases_under()
        assert len(root) == 3
        assert 'cat1' in root
        assert 'cat2' in root
        assert 'cat3' in root

        under_cat1 = get_category_aliases_under('cat1')
        assert len(under_cat1) == 1
        assert 'cat11' in under_cat1


class TestCategoryListBasic(object):

    @pytest.fixture
    def setup(self, user, create_category):
        self.user = user

        self.cat1 = create_category(alias='cat1', creator=self.user, note='some_note')

        self.cl = CategoryList()
        self.cl_cat1 = CategoryList('cat1', show_title=True, show_links=False, cat_html_class='somecss')

    def test_all(self, setup):
        assert str(self.cl) == ''
        assert str(self.cl_cat1) == 'cat1'

        # test_cat_html_class
        assert self.cl.cat_html_class == ''
        assert self.cl_cat1.cat_html_class == 'somecss'

        # test_show_links
        assert self.cl.show_links
        assert self.cl_cat1.show_links == False

        # test_show_title
        assert not self.cl.show_title
        assert self.cl_cat1.show_title == True

        # test_alias
        assert self.cl.alias is None
        assert self.cl_cat1.alias == 'cat1'

        # test_get_category_model
        assert self.cl.get_category_model() is None
        assert self.cl_cat1.get_category_model() == self.cat1

        # test_get_title
        assert self.cl.get_title() == 'Categories'
        assert self.cl_cat1.get_title() == self.cat1.title

        # test_get_id
        assert self.cl.get_id() is None
        assert self.cl_cat1.get_id() == self.cat1.id

        # test_get_note
        assert self.cl.get_note() == ''
        assert self.cl_cat1.get_note() == self.cat1.note


class TestCategoryListNoObj(object):

    @pytest.fixture
    def setup(self, user, create_category):
        self.user = user

        self.cat1 = create_category(alias='cat1', creator=self.user, note='some_note')
        self.cat2 = create_category(alias='cat2', creator=self.user)
        self.cat3 = create_category(alias='cat3')

        self.cat11 = create_category(parent=self.cat1, note='subnote')
        self.cat111 = create_category(parent=self.cat1)

        self.cl = CategoryList()
        self.cl_cat1 = CategoryList('cat1', show_title=True, show_links=False, cat_html_class='somecss')

    def test_all(self, setup):
        # test_get_choices
        choices = self.cl.get_choices()
        assert len(choices) == 3
        assert self.cat1 in choices
        assert self.cat2 in choices
        assert self.cat3 in choices

        choices = self.cl_cat1.get_choices()
        assert len(choices) == 2
        assert self.cat11 in choices
        assert self.cat111 in choices

        # test_get_categories
        cats = self.cl.get_categories()
        assert len(cats) == 3
        assert self.cat1 in cats
        assert self.cat2 in cats
        assert self.cat3 in cats

        cats = self.cl_cat1.get_categories()
        assert len(cats) == 2
        assert self.cat11 in cats
        assert self.cat111 in cats


class TestCategoryListWithObj(object):

    @pytest.fixture
    def setup(self, user, create_article, create_comment, create_category):
        self.article = create_article()
        self.comment = create_comment()
        self.user = user

        self.cat1 = create_category(alias='cat1', creator=self.user, note='some_note')
        self.cat2 = create_category(alias='cat2', creator=self.user)
        self.cat3 = create_category(alias='cat3')

        self.cat11 = create_category(parent=self.cat1, note='subnote')
        self.cat111 = create_category(parent=self.cat1)

        self.cat22 = create_category(parent=self.cat2)
        self.cat222 = create_category(parent=self.cat2)
        self.cat2222 = create_category(parent=self.cat2)

        self.article.add_to_category(self.cat1, self.user)
        self.article.add_to_category(self.cat11, self.user)
        self.article.add_to_category(self.cat3, self.user)

        self.comment.add_to_category(self.cat1, self.user)
        self.comment.add_to_category(self.cat111, self.user)
        self.comment.add_to_category(self.cat2, self.user)
        self.comment.add_to_category(self.cat3, self.user)

        self.cl = CategoryList()
        self.cl.set_obj(self.article)
        self.cl_cat1 = CategoryList('cat1', show_title=True, show_links=False, cat_html_class='somecss')
        self.cl_cat1.set_obj(self.article)

    def test_all(self, setup):
        # test_get_choices
        choices = self.cl.get_choices()
        assert len(choices) == 3
        assert self.cat1 in choices
        assert self.cat2 in choices
        assert self.cat3 in choices

        choices = self.cl_cat1.get_choices()
        assert len(choices) == 2
        assert self.cat11 in choices
        assert self.cat111 in choices

        # test_get_categories
        cats = self.cl.get_categories()
        assert len(cats) == 2
        assert self.cat1 in cats
        assert self.cat2 not in cats
        assert self.cat3 in cats

        cats = self.cl_cat1.get_categories()
        assert len(cats) == 1
        assert self.cat11 in cats


# TODO CategoryRequestHandler



from django.conf import settings


# Path to a model to be used as a Category.
MODEL_CATEGORY = getattr(settings, 'SITECATS_MODEL_CATEGORY', 'sitecats.Category')

# Path to a model to be used as a category-to-object Tie.
MODEL_TIE = getattr(settings, 'SITECATS_MODEL_TIE', 'sitecats.Tie')

# String returned instead of a category URL if unresolved.
UNRESOLVED_URL_MARKER = getattr(settings, 'SITECATS_UNRESOLVED_URL_MARKER', u'#unresolved')



from django.contrib import admin

from .utils import get_category_model, get_tie_model


try:
    from admirarchy.toolbox import HierarchicalModelAdmin
    MODEL_ADMIN = HierarchicalModelAdmin
except ImportError:
    MODEL_ADMIN = admin.ModelAdmin


class CategoryAdmin(MODEL_ADMIN):
    list_display = ('title', 'alias', 'is_locked', 'status')
    search_fields = ('title', 'alias', 'note', 'creator')
    list_filter = ('time_created', 'status')
    ordering = ('sort_order',)
    date_hierarchy = 'time_created'
    hierarchy = True

    def has_delete_permission(self, request, obj=None):
        if obj is not None and obj.is_locked:
            return False
        return super(CategoryAdmin, self).has_delete_permission(request, obj=obj)

    def get_form(self, request, obj=None, **kwargs):
        if obj is not None:
            if obj.is_locked:
                self.readonly_fields = ('is_locked', 'alias')
            else:
                self.readonly_fields = ()
        return super(CategoryAdmin, self).get_form(request, obj=obj, **kwargs)

admin.site.register(get_category_model(), CategoryAdmin)


class TieAdmin(admin.ModelAdmin):
    list_display = ('category', 'content_type', 'object_id', 'status')
    search_fields = ('object_id', 'content_type', 'creator')
    list_filter = ('time_created', 'status', 'content_type')
    ordering = ('-time_created',)
    date_hierarchy = 'time_created'

admin.site.register(get_tie_model(), TieAdmin)



from django import template, VERSION
from django.conf import settings
from django.template.loader import get_template
from django.template.base import FilterExpression

from ..models import ModelWithCategory
from ..toolbox import CategoryRequestHandler, CategoryList
from ..exceptions import SitecatsConfigurationError


register = template.Library()


_CONTEXT_FLATTEN = VERSION >= (1, 11)


@register.tag
def sitecats_url(parser, token):
    tokens = token.split_contents()
    as_clause = detect_clause(parser, 'as', tokens, as_filter_expr=False)
    target_list = detect_clause(parser, 'using', tokens)
    category = detect_clause(parser, 'for', tokens)

    if category is None or target_list is None:
        raise template.TemplateSyntaxError(
            '`sitecats_url` tag expects the following notation: '
            '{% sitecats_url for my_category using my_categories_list as someurl %}.')

    return sitecats_urlNode(category, target_list, as_clause)


@register.tag
def sitecats_categories(parser, token):
    tokens = token.split_contents()
    use_template = detect_clause(parser, 'template', tokens)
    target_obj = detect_clause(parser, 'from', tokens)

    if target_obj is None:
        raise template.TemplateSyntaxError(
            '`sitecats_categories` tag expects the following notation: '
            '{% sitecats_categories from my_categories_list template "sitecats/my_categories.html" %}.')

    return sitecats_categoriesNode(target_obj, use_template)


class sitecats_urlNode(template.Node):

    def __init__(self, category, target_list, as_var):
        self.as_var = as_var
        self.target_list = target_list
        self.category = category

    def render(self, context):
        resolve = lambda arg: arg.resolve(context) if isinstance(arg, FilterExpression) else arg

        category = resolve(self.category)
        target_obj = resolve(self.target_list)

        url = target_obj.get_category_url(category)

        if not self.as_var:
            return url

        context[self.as_var] = url
        return ''


class sitecats_categoriesNode(template.Node):

    def __init__(self, target_obj, use_template):
        self.use_template = use_template
        self.target_obj = target_obj

    def render(self, context):
        resolve = lambda arg: arg.resolve(context) if isinstance(arg, FilterExpression) else arg

        target_obj = resolve(self.target_obj)
        if isinstance(target_obj, CategoryRequestHandler):
            target_obj = target_obj.get_lists()
        elif isinstance(target_obj, ModelWithCategory):
            target_obj = target_obj.get_category_lists()
        elif isinstance(target_obj, (list, tuple)):  # Simple list of CategoryList items.
            pass
        elif isinstance(target_obj, CategoryList):
            target_obj = (target_obj,)
        else:
            if settings.DEBUG:
                raise SitecatsConfigurationError(
                    '`sitecats_categories` template tag can\'t accept `%s` type '
                    'from `%s` template variable.' % (type(target_obj), self.target_obj))
            return ''  # Silent fall.

        context.push()
        context['sitecats_categories'] = target_obj
        template_path = resolve(self.use_template) or 'sitecats/categories.html'
        contents = get_template(template_path).render(context.flatten() if _CONTEXT_FLATTEN else context)
        context.pop()

        return contents


def detect_clause(parser, clause_name, tokens, as_filter_expr=True):
    """Helper function detects a certain clause in tag tokens list.
    Returns its value.

    """
    if clause_name in tokens:
        t_index = tokens.index(clause_name)
        clause_value = tokens[t_index + 1]
        if as_filter_expr:
            clause_value = parser.compile_filter(clause_value)
        del tokens[t_index:t_index + 2]
    else:
        clause_value = None
    return clause_value



from collections import OrderedDict

from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.core.cache import cache
from django.db.models import signals, Count
from etc.toolbox import get_model_class_from_string

from .settings import MODEL_CATEGORY, MODEL_TIE


def get_category_model():
    """Returns the Category model, set for the project."""
    return get_model_class_from_string(MODEL_CATEGORY)


def get_tie_model():
    """Returns the Tie model, set for the project."""
    return get_model_class_from_string(MODEL_TIE)


_SITECATS_CACHE = None


def get_cache():
    """Returns global cache object.

    :rtype: Cache
    :return: cache object
    """
    global _SITECATS_CACHE

    if _SITECATS_CACHE is None:
        _SITECATS_CACHE = apps.get_app_config('sitecats').get_categories_cache()

    return _SITECATS_CACHE


class Cache(object):

    # Sitecats objects are stored in Django cache for a year (60 * 60 * 24 * 365 = 31536000 sec).
    # Cache is only invalidated on sitecats Category model save/delete.
    CACHE_TIMEOUT = 31536000
    CACHE_ENTRY_NAME = 'sitecats'

    CACHE_NAME_IDS = 'ids'
    CACHE_NAME_ALIASES = 'aliases'
    CACHE_NAME_PARENTS = 'parents'

    def __init__(self):
        self._cache = None
        # Listen for signals from the models.
        category_model = get_category_model()
        signals.post_save.connect(self._cache_empty, sender=category_model)
        signals.post_delete.connect(self._cache_empty, sender=category_model)

    def _cache_init(self):
        """Initializes local cache from Django cache if required."""
        cache_ = cache.get(self.CACHE_ENTRY_NAME)

        if cache_ is None:
            categories = get_category_model().objects.order_by('sort_order')

            ids = {category.id: category for category in categories}
            aliases = {category.alias: category for category in categories if category.alias}

            parent_to_children = OrderedDict()  # Preserve aliases order.
            for category in categories:
                parent_category = ids.get(category.parent_id, False)
                parent_alias = None
                if parent_category:
                    parent_alias = parent_category.alias
                if parent_alias not in parent_to_children:
                    parent_to_children[parent_alias] = []
                parent_to_children[parent_alias].append(category.id)

            cache_ = {
                self.CACHE_NAME_IDS: ids,
                self.CACHE_NAME_PARENTS: parent_to_children,
                self.CACHE_NAME_ALIASES: aliases
            }

            cache.set(self.CACHE_ENTRY_NAME, cache_, self.CACHE_TIMEOUT)

        self._cache = cache_

    def _cache_empty(self, **kwargs):
        """Empties cached sitecats data."""
        self._cache = None
        cache.delete(self.CACHE_ENTRY_NAME)

    ENTIRE_ENTRY_KEY = object()

    def _cache_get_entry(self, entry_name, key=ENTIRE_ENTRY_KEY, default=False):
        """Returns cache entry parameter value by its name.

        :param str entry_name:
        :param str key:
        :param type default:
        :return:
        """
        if key is self.ENTIRE_ENTRY_KEY:
            return self._cache[entry_name]
        return self._cache[entry_name].get(key, default)

    def sort_aliases(self, aliases):
        """Sorts the given aliases list, returns a sorted list.

        :param list aliases:
        :return: sorted aliases list
        """
        self._cache_init()
        if not aliases:
            return aliases
        parent_aliases = self._cache_get_entry(self.CACHE_NAME_PARENTS).keys()
        return [parent_alias for parent_alias in parent_aliases if parent_alias in aliases]

    def get_parents_for(self, child_ids):
        """Returns parent aliases for a list of child IDs.

        :param list child_ids:
        :rtype: set
        :return: a set of parent aliases
        """
        self._cache_init()
        parent_candidates = []
        for parent, children in self._cache_get_entry(self.CACHE_NAME_PARENTS).items():
            if set(children).intersection(child_ids):
                parent_candidates.append(parent)
        return set(parent_candidates)  # Make unique.

    def get_children_for(self, parent_alias=None, only_with_aliases=False):
        """Returns a list with with categories under the given parent.

        :param str|None parent_alias: Parent category alias or None for categories under root
        :param bool only_with_aliases: Flag to return only children with aliases
        :return: a list of category objects
        """
        self._cache_init()
        child_ids = self.get_child_ids(parent_alias)
        if only_with_aliases:
            children = []
            for cid in child_ids:
                category = self.get_category_by_id(cid)
                if category.alias:
                    children.append(category)
            return children
        return [self.get_category_by_id(cid) for cid in child_ids]

    def get_child_ids(self, parent_alias):
        """Returns child IDs of the given parent category

        :param str parent_alias: Parent category alias
        :rtype: list
        :return: a list of child IDs
        """
        self._cache_init()
        return self._cache_get_entry(self.CACHE_NAME_PARENTS, parent_alias, [])

    def get_category_by_alias(self, alias):
        """Returns Category object by its alias.

        :param str alias:
        :rtype: Category|None
        :return: category object
        """
        self._cache_init()
        return self._cache_get_entry(self.CACHE_NAME_ALIASES, alias, None)

    def get_category_by_id(self, cid):
        """Returns Category object by its id.

        :param str cid:
        :rtype: Category
        :return: category object
        """
        self._cache_init()
        return self._cache_get_entry(self.CACHE_NAME_IDS, cid)

    def find_category(self, parent_alias, title):
        """Searches parent category children for the given title (case independent).

        :param str parent_alias:
        :param str title:
        :rtype: Category|None
        :return: None if not found; otherwise - found Category
        """
        found = None
        child_ids = self.get_child_ids(parent_alias)
        for cid in child_ids:
            category = self.get_category_by_id(cid)
            if category.title.lower() == title.lower():
                found = category
                break
        return found

    def get_ties_stats(self, categories, target_model=None):
        """Returns a dict with categories popularity stats.

        :param list categories:
        :param Model|None target_model:
        :return:
        """
        filter_kwargs = {
            'category_id__in': categories
        }
        if target_model is not None:
            is_cls = hasattr(target_model, '__name__')
            if is_cls:
                concrete = False
            else:
                concrete = True
                filter_kwargs['object_id'] = target_model.id
            filter_kwargs['content_type'] = ContentType.objects.get_for_model(
                target_model, for_concrete_model=concrete
            )

        return {
            item['category_id']: item['ties_num'] for item in
            get_tie_model().objects.filter(**filter_kwargs).values('category_id').annotate(ties_num=Count('category'))
        }

    def get_categories(self, parent_aliases=None, target_object=None, tied_only=True):
        """Returns subcategories (or ties if `target_object` is set)
        for the given parent category.

        :param str|None|list parent_aliases:
        :param ModelWithCategory|Model target_object:
        :param bool tied_only: Flag to get only categories with ties. Ties stats are stored in `ties_num` attrs.
        :return: a list of category objects or tie objects extended with information from their categories.
        """

        single_mode = False
        if not isinstance(parent_aliases, list):
            single_mode = parent_aliases
            parent_aliases = [parent_aliases]

        all_children = []
        parents_to_children = OrderedDict()
        for parent_alias in parent_aliases:
            child_ids = self.get_child_ids(parent_alias)
            parents_to_children[parent_alias] = child_ids
            if tied_only:
                all_children.extend(child_ids)

        ties = {}
        if tied_only:
            source = OrderedDict()
            ties = self.get_ties_stats(all_children, target_object)
            for parent_alias, child_ids in parents_to_children.items():
                common = set(ties.keys()).intersection(child_ids)
                if common:
                    source[parent_alias] = common

        else:
            source = parents_to_children

        categories = OrderedDict()
        for parent_alias, child_ids in source.items():
            for cat_id in child_ids:
                cat = self.get_category_by_id(cat_id)
                if tied_only:
                    cat.ties_num = ties.get(cat_id, 0)

                if parent_alias not in categories:
                    categories[parent_alias] = []

                categories[parent_alias].append(cat)

        if single_mode != False:  # sic!
            return categories[single_mode]

        return categories



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings
import sitecats.models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('contenttypes', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Category',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('title', models.CharField(help_text='Category name.', max_length=250, verbose_name='Title')),
                ('alias', sitecats.models.CharFieldNullable(null=True, max_length=80, blank=True, help_text='Short name to address category from application code.', unique=True, verbose_name='Alias')),
                ('is_locked', models.BooleanField(default=False, help_text='Categories addressed from application code are locked, their aliases can not be changed. Such categories can be deleted from application code only.', verbose_name='Locked')),
                ('note', models.TextField(verbose_name='Note', blank=True)),
                ('status', models.IntegerField(db_index=True, null=True, verbose_name='Status', blank=True)),
                ('slug', sitecats.models.CharFieldNullable(max_length=250, unique=True, null=True, verbose_name='Slug', blank=True)),
                ('time_created', models.DateTimeField(auto_now_add=True, verbose_name='Date created')),
                ('time_modified', models.DateTimeField(auto_now=True, verbose_name='Date modified')),
                ('sort_order', models.PositiveIntegerField(default=0, help_text='Item position among other categories under the same parent.', verbose_name='Sort order', db_index=True)),
                ('creator', models.ForeignKey(related_name='category_creators', verbose_name='Creator', to=settings.AUTH_USER_MODEL, on_delete=models.CASCADE)),
                ('parent', models.ForeignKey(related_name='category_parents', blank=True, to='sitecats.Category', help_text='Parent category.', null=True, verbose_name='Parent', on_delete=models.CASCADE)),
            ],
            options={
                'abstract': False,
                'verbose_name': 'Category',
                'verbose_name_plural': 'Categories',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Tie',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('note', models.TextField(verbose_name='Note', blank=True)),
                ('status', models.IntegerField(db_index=True, null=True, verbose_name='Status', blank=True)),
                ('time_created', models.DateTimeField(auto_now_add=True, verbose_name='Date created')),
                ('object_id', models.PositiveIntegerField(verbose_name='Object ID', db_index=True)),
                ('category', models.ForeignKey(related_name='tie_categories', verbose_name='Category', to='sitecats.Category', on_delete=models.CASCADE)),
                ('content_type', models.ForeignKey(related_name='sitecats_tie_tags', verbose_name='Content type', to='contenttypes.ContentType', on_delete=models.CASCADE)),
                ('creator', models.ForeignKey(related_name='tie_creators', verbose_name='Creator', to=settings.AUTH_USER_MODEL, on_delete=models.CASCADE)),
            ],
            options={
                'abstract': False,
                'verbose_name': 'Tie',
                'verbose_name_plural': 'Ties',
            },
            bases=(models.Model,),
        ),
        migrations.AlterUniqueTogether(
            name='category',
            unique_together=set([('title', 'parent')]),
        ),
    ]



from django.core.exceptions import ImproperlyConfigured, SuspiciousOperation
from django.forms.fields import ValidationError


class SitecatsException(Exception):
    """Base sitecats exception."""


class SitecatsConfigurationError(SitecatsException, ImproperlyConfigured):
    """Exception for sitecats configuration errors."""


class SitecatsSecurityException(SitecatsException, SuspiciousOperation):
    """Exception raised by CategoryRequestHandler on suspicious request data."""


class SitecatsNewCategoryException(SitecatsSecurityException):
    """Exception raised by CategoryRequestHandler when a try to add a subcategory into a restricted
    category is detected.

    """


class SitecatsLockedCategoryDelete(SitecatsException):
    """Exception raised when a try to delete() a locked category is detected."""


class SitecatsValidationError(SitecatsException, ValidationError):
    """Exception raised by CategoryRequestHandler on request data validation errors."""




from collections import defaultdict

from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.translation import ugettext_lazy as _

from .exceptions import SitecatsLockedCategoryDelete
from .settings import MODEL_CATEGORY, MODEL_TIE
from .utils import get_tie_model

USER_MODEL = getattr(settings, 'AUTH_USER_MODEL', 'auth.User')


class CharFieldNullable(models.CharField):
    """We use custom char field to put nulls in CategoryBase fields.
    That allows 'unique' param to be handled properly.

    """
    def get_prep_value(self, value):
        if value is not None:
            if value.strip() == '':
                return None
        return self.to_python(value)


class CategoryBase(models.Model):
    """Base class for category models.

    Categories form a hierarchical registry to help structuring site entities.

    Inherit from this model and override SITECATS_MODEL_CATEGORY in settings.py
    to customize model fields and behaviour.

    """
    title = models.CharField(
        _('Title'), max_length=250, help_text=_('Category name.'))
    alias = CharFieldNullable(
        _('Alias'), max_length=80,
        help_text=_('Short name to address category from application code.'), blank=True, null=True, unique=True)
    is_locked = models.BooleanField(
        _('Locked'),
        help_text=_('Categories addressed from application code are locked, their aliases can not be changed. '
                    'Such categories can be deleted from application code only.'), default=False)

    parent = models.ForeignKey(
        'self', related_name='%(class)s_parents', verbose_name=_('Parent'),
        help_text=_('Parent category.'), db_index=True, null=True, blank=True, on_delete=models.CASCADE)
    note = models.TextField(_('Note'), blank=True)

    status = models.IntegerField(_('Status'), null=True, blank=True, db_index=True)
    slug = CharFieldNullable(verbose_name=_('Slug'), unique=True, max_length=250, null=True, blank=True)

    creator = models.ForeignKey(
        USER_MODEL, related_name='%(class)s_creators', verbose_name=_('Creator'), on_delete=models.CASCADE)
    time_created = models.DateTimeField(_('Date created'), auto_now_add=True)
    time_modified = models.DateTimeField(_('Date modified'), editable=False, auto_now=True)

    sort_order = models.PositiveIntegerField(
        _('Sort order'),
        help_text=_('Item position among other categories under the same parent.'), db_index=True, default=0)

    class Meta(object):
        abstract = True
        verbose_name = _('Category')
        verbose_name_plural = _('Categories')
        unique_together = ('title', 'parent')

    @classmethod
    def add(cls, title, creator, parent=None):
        """Creates a category.

        :param str title:
        :param User creator:
        :param Category|None parent:
        :return:
        """
        obj = cls(title=title, creator=creator, parent=parent)
        obj.save()
        return obj

    def delete(self, *args, **kwargs):
        """Overridden to handle `is_locked`.

        :param args:
        :param kwargs:
        :return:
        """
        if self.is_locked:
            raise SitecatsLockedCategoryDelete('Unable to delete locked `%s` category.' % self)
        super(CategoryBase, self).delete(*args, **kwargs)

    def save(self, *args, **kwargs):
        """Overridden to set sort order and sanitize title.

        :param args:
        :param kwargs:
        :return:
        """
        self.title = self.title.strip()
        super(CategoryBase, self).save(*args, **kwargs)
        if self.sort_order == 0:
            self.sort_order = self.id
            self.save()

    def __str__(self):
        alias = ''
        if self.alias:
            alias = ' (%s)' % self.alias
        return '%s%s' % (self.title, alias)


class TieBase(models.Model):
    """Base class for ties models.

    Ties are relations between site entities and categories (see above).

    Inherit from this model and override SITECATS_MODEL_TIE in settings.py
    to customize model fields and behaviour.

    """
    category = models.ForeignKey(
        MODEL_CATEGORY, related_name='%(class)s_categories', verbose_name=_('Category'), on_delete=models.CASCADE)
    note = models.TextField(_('Note'), blank=True)
    status = models.IntegerField(_('Status'), null=True, blank=True, db_index=True)

    creator = models.ForeignKey(
        USER_MODEL, related_name='%(class)s_creators', verbose_name=_('Creator'), on_delete=models.CASCADE)
    time_created = models.DateTimeField(_('Date created'), auto_now_add=True)

    # Here follows a link to an object.
    object_id = models.PositiveIntegerField(verbose_name=_('Object ID'), db_index=True)
    content_type = models.ForeignKey(
        ContentType, verbose_name=_('Content type'), related_name='%(app_label)s_%(class)s_tags',
        on_delete=models.CASCADE)

    linked_object = GenericForeignKey()

    class Meta(object):
        abstract = True
        verbose_name = _('Tie')
        verbose_name_plural = _('Ties')

    def __str__(self):
        return '%s:%s tied to %s' % (self.content_type, self.object_id, self.category)

    @classmethod
    def get_linked_objects(cls, filter_kwargs=None, id_only=False, by_category=False):
        """Returns objects linked to categories in a dictionary indexed by model classes.

        :param dict filter_kwargs: Filter for ties.
        :param bool id_only: If True only IDs of linked objects are returned, otherwise - QuerySets.
        :param bool by_category: If True only linked objects and their models a grouped by categories.

        :rtype: defaultdict
        :return:
        """
        filter_kwargs = filter_kwargs or {}

        if by_category:
            results = defaultdict(lambda: defaultdict(list))
        else:
            results = defaultdict(list)

        map_type_to_model = {}

        rows = cls.objects.filter(**filter_kwargs).select_related(
            'content_type', 'category').only('object_id', 'content_type', 'category')

        for row in rows:
            type_id = row.content_type.id
            if type_id not in map_type_to_model:
                map_type_to_model[type_id] = row.content_type.model_class()

            model = map_type_to_model[type_id]
            if by_category:
                results[row.category][model].append(row.object_id)
            else:
                results[model].append(row.object_id)

        if not id_only:
            # Building up QuerySets.
            if by_category:
                for category, linked in results.items():
                    for model, ids in linked.items():
                        results[category][model] = model._base_manager.filter(pk__in=ids)
            else:
                for model, ids in results.items():
                    results[model] = model._base_manager.filter(pk__in=ids)

        return results


class Category(CategoryBase):
    """Built-in category class. Default functionality."""


class Tie(TieBase):
    """Built-in Tie class. Default functionality."""


class ModelWithCategory(models.Model):
    """Helper class for models with tags.

    Mix in this helper to your model class to be able to categorize model instances.

    """
    categories = GenericRelation(MODEL_TIE)

    class Meta(object):
        abstract = True

    _category_lists_init_kwargs = None
    _category_editor = None

    def set_category_lists_init_kwargs(self, kwa_dict):
        """Sets keyword arguments for category lists which can be spawned
        by get_categories().

        :param dict|None kwa_dict:
        :return:
        """
        self._category_lists_init_kwargs = kwa_dict

    def get_category_lists(self, init_kwargs=None, additional_parents_aliases=None):
        """Returns a list of CategoryList objects, associated with
        this model instance.

        :param dict|None init_kwargs:
        :param list|None additional_parents_aliases:
        :rtype: list|CategoryRequestHandler
        :return:
        """

        if self._category_editor is not None:  # Return editor lists instead of plain lists if it's enabled.
            return self._category_editor.get_lists()

        from .toolbox import get_category_lists
        init_kwargs = init_kwargs or {}

        catlist_kwargs = {}
        if self._category_lists_init_kwargs is not None:
            catlist_kwargs.update(self._category_lists_init_kwargs)
        catlist_kwargs.update(init_kwargs)

        lists = get_category_lists(catlist_kwargs, additional_parents_aliases, obj=self)

        return lists

    def enable_category_lists_editor(self, request, editor_init_kwargs=None, additional_parents_aliases=None,
                                     lists_init_kwargs=None, handler_init_kwargs=None):
        """Enables editor functionality for categories of this object.

        :param Request request: Django request object
        :param dict editor_init_kwargs: Keyword args to initialize category lists editor with.
            See CategoryList.enable_editor()
        :param list additional_parents_aliases: Aliases of categories for editor to render
            even if this object has no tie to them.
        :param dict lists_init_kwargs: Keyword args to initialize CategoryList objects with
        :param dict handler_init_kwargs: Keyword args to initialize CategoryRequestHandler object with
        :return:
        """
        from .toolbox import CategoryRequestHandler
        additional_parents_aliases = additional_parents_aliases or []
        lists_init_kwargs = lists_init_kwargs or {}
        editor_init_kwargs = editor_init_kwargs or {}
        handler_init_kwargs = handler_init_kwargs or {}
        handler = CategoryRequestHandler(request, self, **handler_init_kwargs)
        lists = self.get_category_lists(
            init_kwargs=lists_init_kwargs, additional_parents_aliases=additional_parents_aliases)
        handler.register_lists(lists, lists_init_kwargs=lists_init_kwargs, editor_init_kwargs=editor_init_kwargs)
        self._category_editor = handler  # Set link to handler to mutate get_category_lists() behaviour.
        return handler.listen()

    def add_to_category(self, category, user):
        """Add this model instance to a category.

        :param Category category: Category to add this object to
        :param User user: User heir who adds
        :return:
        """
        init_kwargs = {
            'category': category,
            'creator': user,
            'linked_object': self
        }
        tie = self.categories.model(**init_kwargs)  # That's a model of Tie.
        tie.save()
        return tie

    def remove_from_category(self, category):
        """Removes this object from a given category.

        :param Category category:
        :return:
        """
        ctype = ContentType.objects.get_for_model(self)
        self.categories.model.objects.filter(category=category, content_type=ctype, object_id=self.id).delete()

    @classmethod
    def get_ties_for_categories_qs(cls, categories, user=None, status=None):
        """Returns a QuerySet of Ties for the given categories.

        :param list|Category categories:
        :param User|None user:
        :param int|None status:
        :return:
        """
        if not isinstance(categories, list):
            categories = [categories]

        category_ids = []
        for category in categories:
            if isinstance(category, models.Model):
                category_ids.append(category.id)
            else:
                category_ids.append(category)
        filter_kwargs = {
            'content_type': ContentType.objects.get_for_model(cls, for_concrete_model=False),
            'category_id__in': category_ids
        }
        if user is not None:
            filter_kwargs['creator'] = user
        if status is not None:
            filter_kwargs['status'] = status
        ties = get_tie_model().objects.filter(**filter_kwargs)
        return ties

    @classmethod
    def get_from_category_qs(cls, category):
        """Returns a QuerySet of objects of this type associated with the given category.

        :param Category category:
        :rtype: list
        :return:
        """
        ids = cls.get_ties_for_categories_qs(category).values_list('object_id').distinct()
        filter_kwargs = {'id__in': [i[0] for i in ids]}
        return cls.objects.filter(**filter_kwargs)
